import logging
import contextlib

from typing import Union
from json.decoder import JSONDecodeError

from msgraph import core
from azure import identity

from ppa import ui


from . import exceptions


logging.getLogger("azure").setLevel(logging.WARNING)
logger = logging.getLogger(__name__)


EXCEPTION_MAP = {
    "does not comply with password complexity": lambda action: (
        exceptions.InvalidPassword,
        f"Cannot {action} as the supplied password does not meet complexity requirements.",
    ),
    "insufficient privileges": lambda action: (
        exceptions.InsufficientPrivileges,
        f"Cannot {action} as PPA lacks the relevant API permissions.",
    ),
    "already exists": lambda action: (
        exceptions.AlreadyExists,
        f"Cannot {action} as a user already exists with that user principal name.",
    ),
    "does not exist": lambda action: (exceptions.DoesNotExist, f"Cannot {action} as it does not exist."),
    "one or more added object references already exist": lambda action: (
        exceptions.AlreadyExists,
        f"Cannot {action} as it already exists in the group.",
    ),
    "one or more removed object references do not exist": lambda action: (
        exceptions.DoesNotExist,
        f"Cannot {action} as it is not in the group.",
    ),
    "domain portion of the userPrincipalName property": lambda action: (
        exceptions.InvalidParameter,
        f"Cannot {action} as the domain portion of the supplied user principal name is invalid.",
    ),
}


def _generate_exception_message(status_code: int, request_type: str, url: str, response_body: Union[dict, str]) -> str:
    """
    {
        'error': {
            'code': 'Request_BadRequest',
            'message': 'Another object with the same value for property userPrincipalName already exists.',
            'innerError': {
                'date': '2021-09-28T08:22:40',
                'request-id': '9011063c-2a88-4188-adf4-82ace56f86bc',
                'client-request-id': '5b16cf07-09bf-41bf-bf3b-9bd234ce135b'
            }
        }
    }
    """
    return (
        f"({status_code}) Failed {request_type} request to {url}: "
        f"{response_body.get('error', {}).get('message') or response_body if isinstance(response_body, dict) else response_body}"
    )


def _get_matching_exception(exception_message):
    for key in EXCEPTION_MAP:
        if key.lower() in exception_message.lower():
            return EXCEPTION_MAP[key]


@contextlib.contextmanager
def _exception_handler(action, ignored_exceptions=None):
    try:
        yield
    except exceptions.RequestError as e:
        if exception_generator := _get_matching_exception(str(e)):
            exception, message = exception_generator(action)
            if ignored_exceptions is not None and exception in ignored_exceptions:
                logger.info(f"ignoring reraised exception {exception}")
                return
            ui.output_error(message)
            raise exception(message)
        raise e


def graph_api(func):
    def _execute(client, *args, **kwargs):
        # Only check keys & create a GraphClient if it isn't one already (supports one decorated function calling another)
        if isinstance(client, dict):
            if missing_keys := [k for k in {"tenant_id", "client_id", "client_secret"} if k not in client]:
                message = "Supplied graph client is missing the following required keys: {}".format(", ".join(missing_keys))
                ui.output_error(message)
                raise exceptions.BadClientConfiguration(message)
            return func(core.GraphClient(credential=identity.ClientSecretCredential(**client)), *args, **kwargs)
        return func(client, *args, **kwargs)

    return _execute


def logged_request(status_codes: set):
    def _wrapper(func):
        request_type = func.__name__

        def _execute_with_logging(client, url, json=None):
            logger.info(f"sending {request_type} request to endpoint {url}")
            response = func(client, url) if request_type == "get" else func(client, url, json)
            status_code = response.status_code

            if status_code in status_codes:
                if status_code == 204:
                    # 204 (no content) should just return None.
                    return None
                # The data returned by GET requests is usually nested inside the 'value' key of the
                # response, but it isn't when you get a single object (like users.by_principal_name).
                if request_type == "get":
                    try:
                        return response.json()["value"]
                    except KeyError:
                        logger.debug(
                            "No 'value' dictionary was found in the Graph API response, returning the full response instead"
                        )
                return response.json()

            try:
                response_body = response.json()
            except JSONDecodeError:
                # Sometimes a response to a failed request won't have a JSON body.
                response_body = response.text or ""
            raise exceptions.RequestError(_generate_exception_message(status_code, request_type, url, response_body))

        return _execute_with_logging

    return _wrapper


@logged_request({200})
def get(client, url):
    return client.get(url)


@logged_request({200, 201, 204})
def post(client, url, json):
    return client.post(url, json=json)


@logged_request({204})
def patch(client, url, json):
    return client.patch(url, json=json)


@logged_request({204})
def delete(client, url, json):
    return client.delete(url, json=json)
